import java.util.Vector;
import java.io.*;

/**
 * Actors are vector objects in the VIRTUAL WORLD (OH WOW!)
 */

public class Actor {

    int vertices[][];   // a vector is a FP int[3]: x,y,z
    int normals[][];    // a normal is a FP int[3]: x,y,z
    int faces[][];      // a face is a FP int[10]: VERTICES v1,v2,v3,
                        // VERTEX NORMAL INDICES n1,n2,n3,
                        // FACE NORMAL fx,fy,fz,
                        // COLOR c

    int position[]; // Position of the actor

    int rotX, rotY, rotZ;   // Rotation of the actor (in degrees)

    private final static String ROOTPATH = "/gfx"; // where gfx are located
    
    /**
     * Actor constructor (from mesh).
     */
    public Actor(String gobFileName)
        throws IOException
    {
        position = new int[3];
        load(gobFileName);
    }

    /**
     * Load 3D mesh data from a file.
     *
     * @param filename Input filename (in the ROOTPATH)
     */
    private void load(String filename)
        throws IOException
    {
        InputStream resource = getClass().getResourceAsStream(
                ROOTPATH + "/" + filename);
        DataInputStream in = new DataInputStream(resource);

        // XXX it would be nice to check if the file actually
        // exists

        int currentColor = 0x000000;    // Current color of faces

        Vector tempVtx = new Vector();
        Vector tempFace = new Vector();

        boolean done = false;
        while(!done) {
            byte type;
            try {
                type = in.readByte();
            } catch(Exception e) {
                // EOF or Something went wrong
                break;
            }

            switch(type) {
                case 'v':
                    // Vertex: read a new vertex
                    int[] vertex = readVertex(in);  // read a vertex...
                    tempVtx.addElement(vertex);     // add it to the list
                    break;
                case 'c':
                    // Color: change the current color
                    currentColor = readColor(in);
                    break;
                case 'f':
                    // Face: read new face
                    int[] face = readFace(in);
                    face[9] = currentColor;     // set its color...
                    tempFace.addElement(face);  // add it to the list
                    break;
                case '\n':
                    // Blank line: do nothing
                    break;
                default:
                    // Unknown!  Try to continue to the next.
                    System.out.println("UNKNOWN GOB COMMAND '"+type+"' ("+Integer.toHexString((int)type)+")");
                    break;
            }

            // Repeat for next line...
        }

        // Make arrays
        this.vertices = new int[tempVtx.size()][];
        for(int i = 0; i < tempVtx.size(); i++) {
            vertices[i] = (int[])tempVtx.elementAt(i);
        }
        this.faces = new int[tempFace.size()][];
        for(int i = 0; i < tempFace.size(); i++) {
            faces[i] = (int[])tempFace.elementAt(i);
        }
    }

    /**
     * Parse the rest of the line from a DataInputStream into a vector.
     *
     * @param in input stream
     * @return   vector (int[3]: x, y, z)
     */
    private static int[] readVertex(DataInputStream in)
        throws IOException
    {
        int result[] = new int[3];
        for(int i = 0; i < result.length; i++) {
            result[i] = in.readInt();
        }

        System.out.println("Read vertex [ "+FP.toString(result[0])+"  "+FP.toString(result[1])+"  "+FP.toString(result[2])+" ]\n");

        return result;
    }

    /**
     * Parse the rest of the line from a DataInputStream into a color.
     *
     * @param in input stream
     * @return   color
     */
    private static int readColor(DataInputStream in)
        throws IOException
    {
        return in.readInt();
    }

    /**
     * Parse the rest of the line from a DataInputStream into a face.
     *
     * @param in input stream
     * @return   face (int[10]: v1, v2, v3, ni, nj, nk, fx, fy, fz, color)
     * v1, v2, v3 are vertex indices
     * ni, nj, nk are vertex normal indices
     * fx, fy, fz is the normal of this face
     * color is its color
     */
    private static int[] readFace(DataInputStream in)
        throws IOException
    {
        int result[] = new int[10];
        for(int i = 0; i < result.length - 1; i++) {
            result[i] = in.readInt();
        }

        return result;

    }

    /**
     * A descriptive string for this actor.
     * @return String with some info
     */
    public String toString()
    {
        return new String(faces.length+" polys @["+
                FP.toString(position[0])+","+
                FP.toString(position[1])+","+
                FP.toString(position[2])+"]");
    }

    // XXX testing parser NO J2ME BAD BABY!
//    public static void main(String args[])
//        throws IOException
//    {
//        File test = new File("testfile");
//        FileReader in = new FileReader(test);
//
//        // read out first char
//        in.read();
//        int vertices[] = readVertex(in);
//
//        for(int i = 0; i < vertices.length; i++) {
//            System.out.println("\""+vertices[i]+"\"");
//        }
//    }
}

/* vim: set ts=4 sw=4 sta sts=4 et ai: */
