/**
 * Fixed point (16:16) math routines.
 */

public class FP {

    /**
     * Return the fixed point for a whole number.
     *
     * @param whole the whole number to convert
     * @return      the equivalent fixed-point number
     */
    public static int whole(int whole)
    {
        return (whole << 16);
    }

    /**
     * Return the fixed point for a fraction.
     *
     * @param numerator   the numerator
     * @param denominator the denominator
     * @return            the equivalent fixed-point number
     */
    public static int fraction(int numerator, int denominator)
    {
        int frac = Math.abs(numerator) * 65536 / Math.abs(denominator);

        // Determine sign
        boolean negative = false;
        if(numerator > 0 && denominator < 0)
            negative = true;
        else if(numerator < 0 && denominator > 0)
            negative = true;

        // And twos complementify if necessary
        if(negative)
            return (~frac) + 1;
        else
            return frac;

    }

    /**
     * Multiply two fixed-point numbers.
     * 
     * @param f1 Fixed-point number to multiply
     * @param f2 Fixed-point number to multiply
     * @return   Product of f1 and f2
     */
    public static int multiply(int f1, int f2)
    {
        long p = ((long)f1 * f2) >> 16;
        return (int)p;
    }

    /** Divide two fixed-point numbers.
     *
     * @param d1 Dividend
     * @param d2 Divisor
     * @return   Quotient of d1 / d2
     */
    public static int divide(int d1, int d2)
    {
        long q = ( ( ((long)d1) << 32) / d2) >> 16;
        return (int)q;
    }

    /**
     * Return the sine of an angle.
     *
     * @param x angle in degrees
     * @return  sine of x
     */
    public static int sin(int x)
    {
        // Only use the floor of x.  If this is insufficient, we could
        // do linear interpolation.
        x >>= 16;

        // Get it within [0,359]
        while(x < 0) {
            x += 360;
        }
        while(x > 359) {
            x -= 360;
        }

        if(x <= 90) {
            // [0-90] Ascend table
            return SineTable.val[x];
        } else if(x <= 180) {
            // [91-180] Descend table
            return SineTable.val[180 - x];
        } else if(x <= 270) {
            // [181-270] Negative ascend table
            return 0 - SineTable.val[x - 180];
        } else {
            // [271-359] Negative descend table
            return 0 - SineTable.val[360 - x];
        }
    }

    /**
     * Return the cosine of an angle.
     *
     * @param x angle in degrees
     * @return  cosine of x
     */
    public static int cos(int x)
    {
        return sin(whole(90)-x);
    }

    /**
     * Return the tangent of an angle.
     *
     * @param x angle in degrees
     * @return  tangent of x
     */
    public static int tan(int x)
    {
        return divide(sin(x), cos(x));
    }

    /**
     * Return a rounded integer (32-bit) of a fixed-point number.
     *
     * @param n Number to round
     * @return  Rounded 32-bit integer
     */
    public static int round(int n)
    {
        int result = (n >> 16);
        if((n & 0xFFFF) >= 0x8000)
            result++;
        return result;
    }

    /**
     * Return a string representation of a fixed-point number.
     */
    public static String toString(int num)
    {
        int whole, frac;
        StringBuffer result = new StringBuffer();
        boolean negative = (num < 0);

        whole = num >> 16;
        if(negative)
            whole += 1; // twos complement

        // negative zero ;)
        if(whole == 0 && negative)
            result.append("-");
        result.append(Integer.toString(whole));

        if(negative)
            frac = (~num) & 0x0000FFFF; // twos complement
        else
            frac = num & 0x0000FFFF;
        frac = (frac * 10000 / 65536);

        // Rounding
        if(frac % 10 >= 5)
            frac += 10;
        frac /= 10;

        result.append(".");
        result.append(frac);

        return result.toString();
    }

// DEBUG STATEMENTS
// NOT FOR J2ME, BAD BABY!
//
//    /**
//     * Print an FP number, for debugging
//     */
//    public static void debug(int num)
//    {
//        for(int i = 31; i >= 0; i--) {
//            if( ((num >> i) & 1) == 1 )
//                System.out.print("1");
//            else
//                System.out.print("0");
//
//            if(i == 16)
//                System.out.print(" . ");
//        }
//
//        System.out.println(" ~(" + toString(num) + ")");
//    }
//
//    /**
//     * main function, for testing
//     */
//    public static void main(String args[])
//    {
//        System.out.print("sin(45)  = ");
//        debug(sin(whole(45)));
//        System.out.print("cos(45)  = ");
//        debug(cos(whole(45)));
//        System.out.print("tan(45)  = ");
//        debug(tan(whole(45)));
//        System.out.print("sin(-30) = ");
//        debug(sin(whole(-30)));
//        System.out.print("cos(-30) = ");
//        debug(cos(whole(-30)));
//        System.out.print("tan(-30) = ");
//        debug(tan(whole(-30)));
//        System.out.print("-42.27   = ");
//        debug(whole(-42) - fraction(27,100));
//    }
}

/* vim: set ts=4 sw=4 sta sts=4 et ai: */
