import javax.microedition.lcdui.*;
import java.util.Vector;

/**
 * Virtual reality effect (for any vector parts).
 */

public class VirtualReality {
    
    Graphics gfx;       // gfx to use
    Vector actors;      // all actors in the scene

    int camera[];       // camera position (FPI)
    int focus;          // focal length (FPI)

    int xCoef;          // X-coefficient for viewport (FPI)
    int yCoef;          // Y-coefficient for viewport (FPI)

    int xOffset;        // screen x offset (INT)
    int yOffset;        // screen y offset (INT)
    int width;          // screen width (INT)
    int height;         // screen height (INT)
    int centerX;        // X screen center (INT)
    int centerY;        // Y screen center (INT)

    /**
     * Constructor for this class.
     *
     * @param gfx graphics to paint to
     */
    public VirtualReality(Graphics gfx, int xOffset, int yOffset, int width,
            int height)
    {
        this.gfx = gfx;
        this.xOffset = xOffset;
        this.yOffset = yOffset;
        this.width = width;
        this.height = height;
        this.centerX = width/2;
        this.centerY = height/2;

        this.xCoef = FP.whole(1);
        this.yCoef = FP.whole(1);

        setFOV(80);     // default FOV in degrees

        System.out.println("New VR: "+width+"x"+height);

        actors = new Vector();
        camera = new int[3];
    }

    /**
     * Load the data for this scene.
     *
     * @return true if load was successful
     */
    public boolean load()
    {
        // Load up the data
        Actor cube;

        try {
            cube = new Actor("cube.gob");
            actors.addElement(cube);
        } catch(Exception e) {
            return false;
        }
        
        // XXX load real scene
        camera[0] = FP.whole(0);
        camera[1] = FP.whole(0);
        camera[2] = FP.whole(0);
        cube.position[0] = FP.whole(0);
        cube.position[1] = FP.whole(0);
        cube.position[2] = FP.whole(10);

        this.setVirtualDimensions(10, 10);

        // XXX debug
        System.out.println("CUBE: "+cube);

        return true;
    }

    /**
     * Paint the scene to graphics memory.
     */
    public void paint()
    {
        // Translation amount
        int xlate[] = new int[3];
        // Working vectors
        int v1[] = new int[3];
        int v2[] = new int[3];
        int v3[] = new int[3];
        int vn[] = new int[3];
        // Rotation coefficients
        int mx1, my1, mz1, mx2, my2, mz2, mx3, my3, mz3;
        // Final 2D triangle (x1,y1,x2,y2,x3,y3)
        int tri2d[] = new int[6];

        // Loop over all actors
        for(int actIdx = 0; actIdx < actors.size(); actIdx++) {
            Actor actor = (Actor)actors.elementAt(actIdx);
            int vertices[][] = actor.vertices;
            int faces[][] = actor.faces;

            // Translate by camera + actor position
            xlate[0] = camera[0] + actor.position[0];
            xlate[1] = camera[1] + actor.position[1];
            xlate[2] = camera[2] + actor.position[2];

            // Calculate some coefficients for the rotation of this actor
            mx1 = FP.multiply(FP.multiply(FP.sin(actor.rotY),FP.sin(actor.rotX)),FP.sin(actor.rotZ)) + FP.multiply(FP.cos(actor.rotY),FP.cos(actor.rotZ));
            my1 = FP.multiply(FP.cos(actor.rotX),FP.sin(actor.rotZ));
            mz1 = FP.multiply(FP.sin(actor.rotY),FP.cos(actor.rotZ)) - FP.multiply(FP.multiply(FP.cos(actor.rotY),FP.sin(actor.rotX)),FP.sin(actor.rotZ));
            mx2 = FP.multiply(FP.multiply(FP.sin(actor.rotY),FP.sin(actor.rotX)),FP.cos(actor.rotZ)) - FP.multiply(FP.cos(actor.rotY),FP.sin(actor.rotZ));
            my2 = FP.multiply(FP.cos(actor.rotX),FP.cos(actor.rotZ));
            mz2 = 0 - FP.multiply(FP.multiply(FP.cos(actor.rotY),FP.sin(actor.rotX)),FP.cos(actor.rotZ)) - FP.multiply(FP.sin(actor.rotY),FP.sin(actor.rotZ));
            mx3 = 0 - FP.multiply(FP.sin(actor.rotY),FP.cos(actor.rotX));
            my3 = FP.sin(actor.rotX);
            mz3 = FP.multiply(FP.cos(actor.rotY), FP.cos(actor.rotX));

            // Loop over all faces
            for(int facIdx = 0; facIdx < faces.length; facIdx++) {
                int face[] = faces[facIdx];
                int u1[] = vertices[face[0]];
                int u2[] = vertices[face[1]];
                int u3[] = vertices[face[2]];

                // ----
                // Draw
                // ----
                
                // Rotate this thing while it's still on its own axis
                
                v1[0] = FP.multiply(  u1[0]  , mx1) +
                        FP.multiply(  u1[1]  , my1) +
                        FP.multiply(  u1[2]  , mz1);
                v2[0] = FP.multiply(  u2[0]  , mx1) +
                        FP.multiply(  u2[1]  , my1) +
                        FP.multiply(  u2[2]  , mz1);
                v3[0] = FP.multiply(  u3[0]  , mx1) +
                        FP.multiply(  u3[1]  , my1) +
                        FP.multiply(  u3[2]  , mz1);
                vn[0] = FP.multiply(face[0+6], mx1) +   // face normal
                        FP.multiply(face[1+6], my1) +
                        FP.multiply(face[2+6], mz1);
                v1[1] = FP.multiply(  u1[0]  , mx2) +
                        FP.multiply(  u1[1]  , my2) +
                        FP.multiply(  u1[2]  , mz2);
                v2[1] = FP.multiply(  u2[0]  , mx2) +
                        FP.multiply(  u2[1]  , my2) +
                        FP.multiply(  u2[2]  , mz2);
                v3[1] = FP.multiply(  u3[0]  , mx2) +
                        FP.multiply(  u3[1]  , my2) +
                        FP.multiply(  u3[2]  , mz2);
                vn[1] = FP.multiply(face[0+6], mx2) +
                        FP.multiply(face[1+6], my2) +
                        FP.multiply(face[2+6], mz2);
                v1[2] = FP.multiply(  u1[0]  , mx3) +
                        FP.multiply(  u1[1]  , my3) +
                        FP.multiply(  u1[2]  , mz3);
                v2[2] = FP.multiply(  u2[0]  , mx3) +
                        FP.multiply(  u2[1]  , my3) +
                        FP.multiply(  u2[2]  , mz3);
                v3[2] = FP.multiply(  u3[0]  , mx3) +
                        FP.multiply(  u3[1]  , my3) +
                        FP.multiply(  u3[2]  , mz3);
                vn[2] = FP.multiply(face[0+6], mx3) +
                        FP.multiply(face[1+6], my3) +
                        FP.multiply(face[2+6], mz3);

                // Perspective transformation
                // XXX optimization: could focus be a power of 2?
                tri2d[0] = FP.divide(FP.multiply(v1[0]+xlate[0], focus),
                                   v1[2]+xlate[2]);
                tri2d[1] = FP.divide(FP.multiply(v1[1]+xlate[1], focus),
                                   v1[2]+xlate[2]);
                tri2d[2] = FP.divide(FP.multiply(v2[0]+xlate[0], focus),
                                   v2[2]+xlate[2]);
                tri2d[3] = FP.divide(FP.multiply(v2[1]+xlate[1], focus),
                                   v2[2]+xlate[2]);
                tri2d[4] = FP.divide(FP.multiply(v3[0]+xlate[0], focus),
                                   v3[2]+xlate[2]);
                tri2d[5] = FP.divide(FP.multiply(v3[1]+xlate[1], focus),
                                   v3[2]+xlate[2]);

                if(FP.multiply(tri2d[0]-tri2d[2],tri2d[5]-tri2d[3]) - FP.multiply(tri2d[1]-tri2d[3], tri2d[4]-tri2d[2]) > 0) {
                    // Triangle is back-facing
                    // (Check > 0 because our y-axis is INVERTED!)
                    continue;
                }

/*
                System.out.println("FPI: "+
                        FP.toString(x1)+","+FP.toString(y1)+" "+
                        FP.toString(x2)+","+FP.toString(y2)+" "+
                        FP.toString(x3)+","+FP.toString(y3));
*/

                // Apply virtual dimensions
                tri2d[0] = FP.multiply(tri2d[0], xCoef);
                tri2d[1] = FP.multiply(tri2d[1], xCoef);
                tri2d[2] = FP.multiply(tri2d[2], xCoef);
                tri2d[3] = FP.multiply(tri2d[3], xCoef);
                tri2d[4] = FP.multiply(tri2d[4], xCoef);
                tri2d[5] = FP.multiply(tri2d[5], xCoef);

                // Round to whole numbers, and offset to screen
                // XXX optimization: use java graphics.translate()
                tri2d[0] = xOffset + centerX + FP.round(tri2d[0]);
                tri2d[1] = yOffset - centerY + height - FP.round(tri2d[1]);
                tri2d[2] = xOffset + centerX + FP.round(tri2d[2]);
                tri2d[3] = yOffset - centerY + height - FP.round(tri2d[3]);
                tri2d[4] = xOffset + centerX + FP.round(tri2d[4]);
                tri2d[5] = yOffset - centerY + height - FP.round(tri2d[5]);

                gfx.setColor(face[9]);
                gfx.drawLine(tri2d[0], tri2d[1], tri2d[2], tri2d[3]);
                gfx.drawLine(tri2d[2], tri2d[3], tri2d[4], tri2d[5]);
                gfx.drawLine(tri2d[4], tri2d[5], tri2d[0], tri2d[1]);
//                scanTri2D(tri2d);

                /*
                System.out.println("INT: "+
                        x1+","+y1+" "+x2+","+y2+" "+x3+","+y3);
                */
            }
        }
    }

    /**
     * Set the virtual width and height in Vector Units.
     * @param width width of the screen
     * @param height height of the screen
     */
    public void setVirtualDimensions(int width, int height)
    {
        xCoef = FP.divide(FP.whole(this.width), FP.whole(width));
        yCoef = FP.divide(FP.whole(this.height), FP.whole(height));
    }

    /**
     * Set the field of view in degrees.
     * @param degrees Specifies how wide the fov will be
     */
    public void setFOV(int degrees)
    {
        // Find the focal length
        focus = FP.divide(FP.whole(width),
                FP.multiply(FP.whole(2), FP.tan(FP.whole(degrees))));
    }

    /**
     * Scanline convert a triangle to the screen.
     *
     * @param tri[] An int[6] of the coordinates (x1,y1,x2...).  Whole ints,
     * not fixed-point.
     */
    private void scanTri2D(int tri[])
    {

    }
}

/* vim: set ts=4 sw=4 sta sts=4 et ai: */
