#!/usr/bin/perl
#
# Inline fixed-point operations

use POSIX;
use strict;

sub make_fixed {
	my $num = shift;
	my $whole = POSIX::floor($num);
	my $fract = ($num - $whole) * 65536;

	return (int($whole) << 16) + int($fract);
}

# Return the number of parens, or -1 if unbalanced.
sub paren_balance
{
	my $line = shift;
	my $commentblock = shift;
	# Handle comment blocks
	# XXX this will choke on comment blocks within ""s or //s
	if($$commentblock) {
		if($line =~ s/^([^*]|(\*[^\\]))*\*\//*\//) {
			# Found end of comment block.
			$$commentblock = 0;
		} else {
			# Entire line is within comment block: ignore.
			return 0;
		}
	} elsif(($line =~ /\/\*/) && !($line =~ /\/\*([^*]|(\*[^\\]))*\*\//)) {
		# Found beginning of comment block.
		$line =~ s/\/\*.*/\/*/;
		$$commentblock = 1;
	}
	# Ignore parens in quotes and comments
	while($line =~ s/"([^"]*)[()]+([^"]*)"/"$1$2"/g){};
	while($line =~ s/\/\/(.*)[()]+(.*)\n/\/\/$1$2\n/g){};
	my $temp = $line;
	my ($open, $close);
	$open = ($temp =~ s/\(//g);	# only way to count matches?
	$temp = $line;
	$close = ($temp =~ s/\)//g);

	if($open == $close) {
		if($open eq "") { $open = 0; }
		return $open;
	} else {
		return -1;
	}
}

sub parse_args
{
	my ($string, $name, $count) = @_;
	my @args = ();
	my $depth = 0;
	my $index = 0;
	my $c;
	
	# Clean up extra newlines.
	$string =~ s/\n//s;

	# Iterate over every character to find args at depth 0
	foreach $c (split('', $string)) {
		if($c eq '(') { $depth++; }
		if($c eq ')') { $depth--; }
		if($c eq ',' && $depth == 0) {
			$index++;
		} else {
			$args[$index] .= $c;
		}
	}

	# Check the count.
	if($#args != ($count - 1)) {
		print "WRONG # OF ARGUMENTS TO $name: ". ($#args+1) . " should be $count\n";
		exit 1;
	}

	return \@args;
}

# Optimize a line or set of lines.
sub optimize
{
	my $line = shift;

	# This is from Mastering Regular Expressions chapter 4
	my $nesting = 8;	# maximum nesting level
	my $paren_match = '(' . '([^()]|\(' x $nesting .
			  '[^()]*' . '\))*' x $nesting . ')';

	# Run multiple passes until all optimizes complete
	my $matches;
	do {
		$matches = 0;
		if($line =~ s/FP\.(multiply|divide)\s*\($paren_match\)/XXX^REPL/s) {
			my $operation = $1;
			my @args = @{parse_args($2, $operation, 2)};

			# Replace a single mult/div op
			if($operation eq 'multiply') {
				$line =~ s/XXX\^REPL/((int)(((long)($args[0])*($args[1])) >> 16))/;
			} elsif($operation eq 'divide') {
				$line =~ s/XXX\^REPL/((int)(((((long)($args[0]))<<32)\/($args[1])) >> 16))/;
			}
			$matches = 1;
		}

		if($line =~ s/FP\.whole\s*\($paren_match\)/XXX^REPL/s) {
			my @args = @{parse_args($1, "whole", 1)};

			# Plain number
			if($args[0] =~ m/^\s*(-?\d+)\s*$/) {
				my $fpi = sprintf("(0x%X)", make_fixed($1));
				$line =~ s/XXX\^REPL/$fpi/;
			} else {
				$line =~ s/XXX\^REPL/(($args[0]) << 16)/;
			}
			$matches = 1;
		}

		my $undo_fraction = $line;
		if($line =~ s/FP\.fraction\s*\($paren_match\)/XXX^REPL/s) {
			my @args = @{parse_args($1, "fraction", 2)};

			# Plain numbers
			my $plain_ok = 0;
			if($args[0] =~ m/^\s*(-?\d+)\s*$/) {
				my $num1 = $1;
				if($args[1] =~ m/^\s*(-?\d+)\s*$/) {
					my $num2 = $1;
					my $frac = 1.0*$num1/$num2;
					my $fpi = sprintf("(0x%X)", make_fixed($frac));
					$line =~ s/XXX\^REPL/$fpi/;
					$plain_ok = 1;
				}
			}
			if(!$plain_ok) {
				# Couldn't do anything.
				$line = $undo_fraction;
			}
		}
			
	} while($matches != 0);

	print $line;
}


# Loop over every line
my ($line, $newline);
my $commentblock = 0;
while(($line = <>)) {

	# Get more lines if the expression is incomplete.
	while(paren_balance($line, \$commentblock) < 0) {
		if(!($newline = <>)) {
			print "ERROR: Expression incomplete at EOF\n";
			exit 1;
		}
		chop($line);
		$newline =~ s/^(\s*)/ /;
		$line .= $newline;
	}

	optimize($line);
}
