#!/usr/bin/perl
#
# Pack an OBJ file into our binary GOB format.

use POSIX;
use strict;

sub make_fixed {
	my $num = shift;
	my $whole = POSIX::floor($num);
	my $fract = ($num - $whole) * 65536;

	return (int($whole) << 16) + int($fract);
}

sub crossTri {
	my ($a, $b, $c) = @_;
	my $d = [];

	my $v = [$$b[0]-$$a[0], $$b[1]-$$a[1], $$b[2]-$$a[2]];
	my $w = [$$c[0]-$$a[0], $$c[1]-$$a[1], $$c[2]-$$a[2]];
	$$d[0] = $$v[1]*$$w[2] - $$v[2]*$$w[1];
	$$d[1] = $$v[2]*$$w[0] - $$v[0]*$$w[2];
	$$d[2] = $$v[0]*$$w[1] - $$v[1]*$$w[0];
	return $d;
}

sub normalize {
	my $v = shift;
	my $w = [];
	my $m = sqrt($$v[0]*$$v[0]+$$v[1]*$$v[1]+$$v[2]*$$v[2]);
	$$w[0] = $$v[0]/$m;
	$$w[1] = $$v[1]/$m;
	$$w[2] = $$v[2]/$m;
	return $w;
}

# Check args
if($#ARGV != 1) {
	print "USAGE: gob_pack.pl <input.obj> <output.gob>\n";
	exit 0;
}

# Prepare files
my ($infile, $outfile);
if(!open($infile, '<', $ARGV[0])) {
	print "Can't open OBJ file ".$ARGV[0]."\n";
	exit 1;
}
if(!open($outfile, '>', $ARGV[1])) {
	print "Can't create GOB file ".$ARGV[1]."\n";
	exit 2;
}

# Data we care about
my @vertices = ();	# (x,y,z)
my @normals = ();	# (x,y,z)
my @faces = ();		# (vi,vj,vk,ni,nj,nk,fnx,fny,fnz)
			# n[ijk] is the index of the normals
			# fn[xyz] is the normal of the whole face
my %colors = ();	# {faceIdx: color}

# Loop over every line in the input file and load all data.
while(my $line = <$infile>) {
	# Vertex
	if($line =~ m/^\s*v\s+([\d.-]+)\s+([\d.-]+)\s([\d.-]+)\s*$/) {
		push @vertices, [$1, $2, $3];
		next;
	}
	# Face
	if($line =~ m/^\s*f\s+(\d+)[ \t]*\/[ \t]*(\d+)[ \t]*\/[ \t]*(\d+)[ \t]*([ \t]+(\d+)[ \t]*\/[ \t]*(\d+)[ \t]*\/[ \t]*(\d+)[ \t]*)?$/) {
		my @face = ();
		if(defined $4) {
			# Normals given
			@face = ($1-1, $2-1, $3-1,
				 $5-1, $6-1, $7-1,
				 undef, undef, undef);
		} else {
			# Calc normals later
			@face = ($1-1, $2-1, $3-1,
				 undef, undef, undef,
				 undef, undef, undef);
		}
		push @faces, \@face;
		next;
	}
	# Color
	if($line =~ m/^\s*g\s+([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})\s*$/i) {
		$colors{$#faces + 1} = [$1, $2, $3];
		next;
	}
	# Blank lines/comments
	if($line =~ m/^(\s*|#.*)$/) {
		next;
	}

	# WHA?!
	print " * WARNING!  Skipping unknown line:\n".$line;
}

# Calculate all face normals
my ($f, $i, $normal);
foreach $f (@faces) {
	$normal = normalize(crossTri($vertices[$$f[0]],
			$vertices[$$f[1]], $vertices[$$f[2]]));
	$$f[6] = $$normal[0];
	$$f[7] = $$normal[1];
	$$f[8] = $$normal[2];
	# XXX actually calculate these
	$$f[3] = 0;
	$$f[4] = 0;
	$$f[5] = 0;
	# out of this loop, into another to calculate vertex normals
}

my $v;
foreach $v (@vertices) {
	$v = [map {make_fixed($_)} @$v];
	print {$outfile} pack("CNNN", ord("v"), @$v);
}

my $i;
for($i = 0; $i <= $#faces; $i++) {
	# Check for color change.
	my $c = $colors{$i};
	if(defined $c) {
		print {$outfile} pack("CCCCC", ord("c"), 0,
			hex($$c[0]), hex($$c[1]), hex($$c[2]));
	}

	$f = $faces[$i];
	# Make the face normal fixed-point
	@$f[6..8] = map {make_fixed($_)} @$f[6..8];
	print {$outfile} pack("CNNNNNNNNN", ord("f"), @$f);
}


print "Packed ".($#vertices + 1)." vertices, ".($#faces + 1)." faces, and ".
	keys(%colors)." color changes.\n";

# Clean up
close($infile);
close($outfile);
